unit ChunkedPostFix;
                   
{ content parser for chunked POST queries.

 This unit parses all web requests of type "application/x-www-form-urlencoded".
 Normally these requests are handled by the standard HTTP parser but a bug
 (if it is one) in the implementation causes the parser not to load the entire
 content postet (because ReadClient does not read all the data requested).

 So this content parser accepts all URL encoded POST queries avoiding this
 problem. A timer takes care that the process does not wait indefinitely for
 data when the user breaks the upload process by hitting STOP.

 Note this is designed for CGI type applications, but should work for ISAPI,
 NSAPI or Apache DSO as well.

 WARNING: Because this is only a workaround, the variable Request.Content may
          not contain all the post data. This is because when the fix becomes
          active, the additional POST data is read into an internal buffer and
          there is no way to alter the "Content" property from outside the
          object. Note also that the function ExtractContentFields is not
          used. Instead, it is implemented internally (however, ExtractFields
          is still used).

 Copyright by Udo Giacomozzi <udo@nova-sys.net>
 Version: 1.0

 www.nova-sys.net
}

interface

uses
  Classes, SysUtils, DateUtils, HTTPApp;

type
  EContentTooLarge    = class(Exception);
  EContentReadTimeout = class(Exception);

  TChunkedContentParser = class(TAbstractContentParser)
  private
    FContentFields: TStrings;
  public
    function GetContentFields: TStrings; override;
    function GetFiles: TAbstractWebRequestFiles; override;
    class function CanParse(AWebRequest: TWebRequest): Boolean; override;
  end;



implementation

const
  RequestTimeout = 60*1000;
    { Timeout, in ms, after an exception is raised if not the entire request
      content could be read. TChunkedContentParser tries to read the POST
      content bit by bit (or, chunk by chunk, if you want) until the size
      specified in ContentLength has been read. The timeout represents the
      time from the last byte received, so posting large amount of data with
      slow connections is no problem. For example, if RequestTimeout is
      60000, an exception is rised if for about 60 seconds no additional
      byte has been received. }

  MaxPostSize = 48*1024;
    { Maximim size, in bytes, that should be read. If this is greater than 0,
      the content parser will raise an exception when the ContentLength is
      greater than MaxPostSize. You can set any value but remember that all
      data is loaded into memory. }

  SleepTime = 1;
    { Sleep time, in ms, that the parser waits when no data could be read.
      This reduces server load for very slow connections or when the user
      interrupts the POST process. }


class function TChunkedContentParser.CanParse(AWebRequest: TWebRequest): Boolean;
begin
  // Use this only for FORM POST data, not for files
  Result := UpperCase(AWebRequest.ContentType) =
            'APPLICATION/X-WWW-FORM-URLENCODED';
end;

function TChunkedContentParser.GetContentFields: TStrings;
var
  MyContent          : AnsiString;
  NewBlock           : AnsiString;
  LastBlockTime      : TDateTime;
begin
  if FContentFields = nil then
  begin
    FContentFields := TStringList.Create;

    // Content-Length must be specified
    If WebRequest.ContentLength > 0 then
    begin

      // Copy already loaded content
      MyContent := WebRequest.Content;

      // Load any missing content
      LastBlockTime:=Now;
      If WebRequest.ContentLength > MaxPostSize then
        raise EContentTooLarge.Create('POST content too large');

      while (Length(MyContent) < WebRequest.ContentLength)
         and (MilliSecondsBetween(Now, LastBlockTime) < RequestTimeout) do
      begin
        // Try to read missing data - note that it is possible that
        // less data is read than requested (this is/was in fact the central
        // problem...)
        NewBlock :=
          WebRequest.ReadString(WebRequest.ContentLength-Length(MyContent));

        If NewBlock<>'' then
        begin
          MyContent := MyContent + NewBlock;
          LastBlockTime:=Now;
        end else Sleep(SleepTime); // reduce cpu usage
      end;

      // Check loaded data (timeout?)
      If Length(MyContent) < WebRequest.ContentLength then
      begin
        // not all data could be loaded -> raise exception
        raise EContentReadTimeout.Create('Timeout reading POST data');
      end;

      // Parse content like ExtractContentFields does
      WebRequest.ExtractFields(['&'], [], PChar(MyContent), FContentFields);
    end;

  end;
  Result := FContentFields;
end;



// Like the standard TContentParser we don't support files
// For file uploading another Content-Type is used so you can use
// a multipart form parser along with this unit without problems.
type
  TMyEmptyRequestFiles = class(TAbstractWebRequestFiles)
  protected
    function GetCount: Integer; override;
    function GetItem(I: Integer): TAbstractWebRequestFile; override;
  end;

function TMyEmptyRequestFiles.GetCount: Integer;
begin
  Result := 0;
end;

function TMyEmptyRequestFiles.GetItem(I: Integer): TAbstractWebRequestFile;
begin
  Result := nil;
end;

function TChunkedContentParser.GetFiles: TAbstractWebRequestFiles;
begin
  Result := TMyEmptyRequestFiles.Create;
end;


initialization

  RegisterContentParser(TChunkedContentParser);

end.
